"""Api class for iNels BUS."""
import logging

from functools import partial
from pyinels.device import Device

from pyinels.const import (
    NAME,
    VERSION,
    ATTR_DOWN,
    ATTR_GROUP,
    ATTR_ID,
    ATTR_UNKNOWN,
    ATTR_SHUTTER,
    ATTR_TEMP,
    ATTR_THERM,
    ATTR_TITLE,
    ATTR_TYPE,
    ATTR_UP,
    DEVICE_TYPE_DICT,
    INELS_BUS_ATTR_DICT
)

from pyinels.exception import (
    ApiConnectionException,
    ApiDataTypeException,
    ApiException
)


from xmlrpc.client import ServerProxy

_LOGGER = logging.getLogger(__name__)

class Api:
    """Class of iNels BUS."""

    def __init__(self, host, port, version):
        """Initialize Api class."""
        self.__host = host
        self.__port = port
        self.__version = version
        self.__proxy = None
        self.__devices = None

        self.package_name = NAME
        self.package_version = VERSION

    @property
    def proxy(self):
        """Proxy of the bus server."""
        if isinstance(self._Api__proxy, ServerProxy):
            return self.__proxy

        self.__proxy = self.__conn()
        return self.__proxy

    @property
    def devices(self):
        """Loaded devices by getAllDevices."""
        if self.__devices is None:
            self.__devices = self.getAllDevices()

        return self.__devices

    def set_devices(self, devices):
        """Set device prop."""
        self.__devices = devices

    def __conn(self):
        """Instantient Api iNels BUS connection class."""
        try:
            con = ServerProxy(self.__host + ":" + str(self.__port))
            return con
        except BlockingIOError as err:
            raise ApiConnectionException(err.errno, err.strerror, err)
        except Exception as err:
            raise ApiException(
                "common_exception", "Exception occur", err)

    def ping(self):
        """Check connection iNels BUS with ping."""
        return self.proxy.ping()

    def get_package_name(self):
        return self.package_name

    def get_package_version(self):
        return self.package_version

    def getPlcIp(self):
        """Get Ip address of PLC."""
        return self.proxy.getPlcIP()

    def getRooms(self):
        """List of all rooms from Connection server website."""
        return self.proxy.getRooms()

    def getRoomDevicesRaw(self, room_name):
        """List of all devices in given room."""
        return self.proxy.getRoomDevices(room_name)

    def getRoomDevices(self, room_name):
        """List of all devices in defined room."""
        return self.__roomDevicesToJson(room_name)

    def getAllDevices(self):
        """Get all devices from all rooms."""
        devices = []

        rooms = self.getRooms()
        # rooms = ['fürdőszoba']
        # go trough all rooms
        for room in rooms:
            # print('ROOM:', room)
            room_devices = self.getRoomDevices(room)
            # go trough all devices in room
            for room_dev in room_devices:
                is_in = False
                # check when duplicate devices occured
                for device in devices:
                    if device.id == room_dev.id:
                        is_in = True
                        break
                # not presented in the list, then append
                if is_in is False:
                    devices.append(room_dev)

        self.set_devices(devices)

        return devices

    def fetch_all_devices(self):
        """Fetch all devices data."""
        if self.__devices is not None and len(self.__devices) > 0:
            dev_ids = []

            for dev in self.__devices:
                # shutters has composed ids, and inels dont have any
                # for them so we need to use up and down identifiers instead
                if dev.up is not None and dev.down is not None:
                    dev_ids.append(dev.up)
                    dev_ids.append(dev.down)
                else:
                    dev_ids.append(dev.id)

            data_values = self.read(dev_ids)

            # rerender data on all devices
            for dev in self.__devices:
                if dev.up is not None and dev.down is not None:
                    dev.set_value({
                        f'{dev.down}': data_values[dev.down],
                        f'{dev.up}': data_values[dev.up]
                    })
                else:
                    dev.set_value({f'{dev.id}': f'{data_values[dev.id]}'})

            # return list of ids and actual data states
            return data_values

        return None

    def read(self, device_ids):
        """Get the value from the proxy by device id."""
        if not isinstance(device_ids, list):
            raise ApiDataTypeException(
                'readDeviceData', f'{device_ids} is not a list!')
        return self.__readDeviceData(device_ids)

    def write(self, device, value):
        """Write data to multiple devices."""
        if not hasattr(device, 'id'):
            raise ApiDataTypeException(
                'readDeviceData', f'{device} has no id')
        try:
            self.__writeValues(value)
        except Exception as err:
            raise ApiException("write_proxy", err)

    def __writeValues(self, command):
        """Write data to the proxy."""
        self.proxy.writeValues(command)

    def __roomDevicesToJson(self, room_name):
        """Create json object from devices listed in preffered room."""
        d_type = None
        devices = []

        raw_list = self.getRoomDevicesRaw(room_name)
        devices_list = raw_list.split('\n')

        # TODO - remove
        # print (raw_list)
        # return

        for item in devices_list:
            start = len(item) - 1
            end = len(item)

            # TODO - remove
            print(item)
            if start > 0:
                if item[start:end] == ":":
                    d_type = item[0:start]

                    # Ignore 
                    # if d_type in "blank":
                    #     break
                else:
                    json_dev = item.split('" ')
                    obj = {}
                    obj[INELS_BUS_ATTR_DICT.get(ATTR_GROUP)] = room_name
                    
                    # TODO - remove
                    # print(item)
                    for prop in json_dev:
                        frag = prop.split("=")
                        # TODO - remove
                        print(frag)
                        obj[frag[0]] = frag[1].replace("\"", " ").strip()

                    # TODO - remove
                    # obj[INELS_BUS_ATTR_DICT.get(ATTR_TYPE)] = DEVICE_TYPE_DICT.get(d_type)

                    # TODO - remove
                    # print(f"obj: {obj} , d_type: {d_type}")

                    obj[INELS_BUS_ATTR_DICT
                        .get(ATTR_TYPE)] = DEVICE_TYPE_DICT.get(
                            d_type, "unknown")

                    obj = self.__recognizeAndSetUniqueIdToDevice(obj)

                    device = Device(obj, self)
                    device.get_value()

                    # TODO - remove
                    # print("THIS IS AN ID:", device.id)

                    devices.append(device)

        return devices

    def __recognizeAndSetUniqueIdToDevice(self, raw_device):
        """Some of the devices does not have unique id
        presented in inels attribute. We need do create
        one from other unique attributes."""

        def set_shutter_id(dev):
            """Set the id to the shutter."""
            dev[INELS_BUS_ATTR_DICT.get(
                ATTR_ID)] = dev[INELS_BUS_ATTR_DICT.get(ATTR_UP)] + \
                "_" + dev[INELS_BUS_ATTR_DICT.get(ATTR_DOWN)]

            return dev

        def set_therm_id(dev):
            """Set the id to the therms."""
            dev[INELS_BUS_ATTR_DICT.get(
                ATTR_ID)] = dev[INELS_BUS_ATTR_DICT.get(ATTR_TEMP)]

            return dev

        def set_not_known_id_from_name(dev):
            """Set the id to the not know device from name."""
            name = dev[INELS_BUS_ATTR_DICT.get(ATTR_TITLE)].replace(" ", "_")
            dev[INELS_BUS_ATTR_DICT.get(ATTR_ID)] = name

            return dev

        # use a switch to create identifier inside of the raw data
        # from usefull attributes

        # TODO - remove
        # print("DOES THS GET HIT2")
        # print(raw_device)

        # Todo - this part does not get hit, if it has an inels property
        if INELS_BUS_ATTR_DICT.get(ATTR_ID) not in raw_device:
            switcher = {
                ATTR_SHUTTER: partial(set_shutter_id, raw_device),
                ATTR_THERM: partial(set_therm_id, raw_device),
                ATTR_UNKNOWN: partial(set_not_known_id_from_name, raw_device)
            }

            # print("DOES THS GET HIT")
            # print(raw_device)
            # print(switcher.get(raw_device[INELS_BUS_ATTR_DICT.get(ATTR_TYPE)]))

            
            # print(raw_device[INELS_BUS_ATTR_DICT.get(ATTR_TYPE)])
            fnc = switcher.get(raw_device[INELS_BUS_ATTR_DICT.get(ATTR_TYPE)])
            # call selected function to set the identifier
            raw_device = fnc()

        return raw_device

    def __readDeviceData(self, device_ids):
        """Reading devices data from proxy."""
        return self.proxy.read(device_ids)



"""Production test cases."""
import asyncio

from pyinels.api import Api
from pyinels.device.pyLight import pyLight
from pyinels.device.pySwitch import pySwitch
from pyinels.device.pyDoor import pyDoor
from pyinels.device.pyShutter import pyShutter

# from unittest import async_case

MMS_IP_ADDRESS = "192.168.2.102"
PLC_IP_ADDRESS = "192.168.2.101"


# class ProductionTest(async_case.IsolatedAsyncioTestCase):
class ProductionTest():
    """Library used agains production server."""

    def setUp(self):
        """Setup all necessary instances nad mocks."""
        self.api = Api(f'http://{MMS_IP_ADDRESS}', 8001, "CU3")

    async def asyncSetUp(self):
        """Setup all neccessary async stuff."""
        devices = await self.api.getAllDevices()
        self.api.set_devices(devices)

    def tearDown(self):
        """Remove all attached properties."""
        self.api = None

    async def test_ping_success(self):
        """Ping test."""
        ping = await self.api.ping()

        self.assertEqual(ping, True)

    async def test_plcIp_address(self):
        """Get Ip address of the PLC."""
        ip = await self.api.getPlcIp()

        self.assertEqual(PLC_IP_ADDRESS, ip)

    def test_loaded_devices(self):
        """Are devices from api loaded?"""
        self.assertGreater(len(self.api.devices), 0)

    async def test_create_light(self):
        """create and test light."""
        devices = [
            x for x in self.api.devices
            if x.id == "SV_7_Pokoj_dole"]

        light = await pyLight(devices[0])
        self.assertEqual(light.state, False)

        await light.turn_on()
        self.assertEqual(light.state, True)
        asyncio.sleep(4)
        await light.turn_off()
        self.assertEqual(light.state, False)
        asyncio.sleep(4)

        await light.set_brightness(50)
        self.assertEqual(light.state, True)
        asyncio.sleep(4)
        await light.set_brightness(100)
        self.assertEqual(light.state, True)
        asyncio.sleep(4)
        await light.set_brightness(0)
        self.assertEqual(light.state, False)

    async def test_create_switch(self):
        """create and test switch."""
        devices = [
            x for x in self.api.devices if x.id
            == "ZAS_1B_Pokoj_dole"]

        switch = await pySwitch(devices[0])

        await switch.turn_off()
        self.assertEqual(switch.state, False)
        asyncio.sleep(4)
        await switch.turn_on()
        self.assertEqual(switch.state, True)

    async def test_create_door(self):
        """crate and test door."""
        devices = [
            x for x in self.api.devices if x.id
            == "Vrata_Garaz"]

        door = await pyDoor(devices[0])

        await door.turn_on()
        asyncio.sleep(30)
        await door.turn_off()

    async def test_create_shutter(self):
        """create and test shutter."""
        devices = [
            x for x in self.api.devices
            if x.id == "ROL_Pokoj_host_nahoru_ROL_Pokoj_host_dolu"]

        shutter = await pyShutter(devices[0])

        await shutter.pull_up()
        asyncio.sleep(20)
        await shutter.stop()
        await shutter.pull_down()
        asyncio.sleep(10)
        await shutter.stop()



"""Inels shutter class to control blinds."""
from pyinels.device.pyBase import pyBase
# from pyinels.pyTimer import pyTimer

from pyinels.const import (
    ATTR_DOWN,
    ATTR_UP,
    ATTR_STOP,
    ATTR_SWITCH_ON,
    ATTR_SWITCH_OFF,
    DIRECTIONS_DICT,
    RANGE_BLIND,
    STATE_OPEN,
    SUPPORT_OPEN,
    SUPPORT_CLOSE,
    SUPPORT_SET_POSITION,
    SUPPORT_STOP,
    SUPPORT_OPEN_TILT,
    SUPPORT_CLOSE_TILT,
    SUPPORT_STOP_TILT,
    STATE_CLOSING,
    STATE_OPENING,
    STATE_CLOSED
)

MIN_RANGE = RANGE_BLIND[0]
MAX_RANGE = RANGE_BLIND[1]


class pyShutter(pyBase):
    """Inels class shutter."""

    def __init__(self, device):
        """Initialize shutter."""
        super().__init__(device)
        self._current_possition = 0
        self._previous_state = self._device.value

    @property
    def state(self):
        """State where the shutter is."""
        up_device = self.up
        down_device = self.down

        up_on = up_device == ATTR_SWITCH_ON \
            and down_device == ATTR_SWITCH_OFF
        down_on = up_device == ATTR_SWITCH_OFF \
            and down_device == ATTR_SWITCH_ON

        previous_up = self._previous_state[self._device.up]
        previous_down = self._previous_state[self._device.down]

        prev_up_on = previous_up == ATTR_SWITCH_ON \
            and previous_down == ATTR_SWITCH_OFF
        prev_down_on = previous_up == ATTR_SWITCH_OFF \
            and previous_down == ATTR_SWITCH_ON

        # returning state based on current state of up and down and previous
        # state up and down
        return (STATE_OPENING if up_on and not down_on else STATE_CLOSING
                if not up_on and down_on else STATE_CLOSED if not prev_up_on
                and prev_down_on else STATE_OPEN)

    @property
    def supported_features(self):
        """Definition what the devices supports."""
        return SUPPORT_OPEN \
            | SUPPORT_CLOSE \
            | SUPPORT_SET_POSITION \
            | SUPPORT_STOP \
            | SUPPORT_OPEN_TILT \
            | SUPPORT_CLOSE_TILT \
            | SUPPORT_STOP_TILT

    @property
    def current_position(self) -> int:
        """Current position of the shutter."""
        return self._current_possition

    @property
    def value(self):
        """Overrided value, becuase shutter has two values in one."""
        return self._device.value

    def pull_up(self, stop_after=None):
        """Turn up the shutter."""
        self.__call_service(DIRECTIONS_DICT.get(ATTR_UP))

    def pull_down(self, stop_after=None):
        """ Turn down the shutter."""
        self.__call_service(DIRECTIONS_DICT.get(ATTR_DOWN))

    def stop(self):
        """ Stop the shutter."""
        self.__call_service(DIRECTIONS_DICT.get(ATTR_STOP))

    def __call_service(self, direction):
        """Internal call of the device write value."""
        self._previous_state = self._device.value

        if direction == DIRECTIONS_DICT.get(ATTR_STOP):
            prev_up = self._previous_state[self._device.up]
            prev_down = self._previous_state[self._device.down]
            # when previous up state is
            self._current_possition = (0 if prev_down == 1
                                       and prev_up == 0
                                       else 100
                                       if prev_down == 0
                                       and prev_up == 1
                                       else 0)

            self._device.write_value(self.__set_value(
                ATTR_SWITCH_OFF, ATTR_SWITCH_OFF))
        elif direction == DIRECTIONS_DICT.get(ATTR_UP):
            self._current_possition = 100
            self._device.write_value(self.__set_value(
                ATTR_SWITCH_OFF, ATTR_SWITCH_ON))
        elif direction == DIRECTIONS_DICT.get(ATTR_DOWN):
            self._current_possition = 0
            self._device.write_value(self.__set_value(
                ATTR_SWITCH_ON, ATTR_SWITCH_OFF))

    def __set_value(self, down, up):
        """Set the value to call service."""
        return {f'{self._device.down}': down, f'{self._device.up}': up}



NAME = "pyinels-domenix-dev"
VERSION = "0.2.0"

"""Const of iNels BUS."""

# Inels attributes
ATTR_DOWN = "Down"
ATTR_GROUP = "Group"
ATTR_ID = "Id"
ATTR_READ_ONLY = "Read_only"
ATTR_RELE = "Rele"
ATTR_SWITCH_OFF = 0
ATTR_SWITCH_ON = 1
ATTR_TEMP = "Temp"
ATTR_TEMP_SET = "Temp_set"
ATTR_TITLE = "Title"
ATTR_TYPE = "Type"
ATTR_UP = "Up"
ATTR_STOP = "Stop"

# Internal device attributes
ATTR_SWITCH = "switch"
ATTR_LIGHT = "light"
ATTR_DOOR = "door"
ATTR_SHUTTER = "shutter"
ATTR_HEATING = "heating"
ATTR_THERM = "therm"
ATTR_UNKNOWN = "unknown"

# Device types dictionary with all determined typs translated to better
# understand
DEVICE_TYPE_DICT = {
    "undefined": "undefined",
    "lights": "light",
    "lamps": "light",
    "on_off": "switch",
    "shutters": "shutter",
    "heat-control": "therm",
    "thermals": "therm",
    "heating": "heating",
    "garage": "door",
    "gate": "door"
    # "scenes": "tobecontinued"
}

INELS_BUS_ATTR_DICT = {
    "Id": "inels",
    "Title": "name",
    "Type": "type",
    "Temp": "therm",
    "Rele": "rele",
    "Temp_set": "stateth",
    "Down": "down",
    "Up": "up",
    "Read_only": "read_only",
    "Group": "group",
}

DIRECTIONS_DICT = {
    "Up": "up",
    "Down": "down",
    "Stop": "stop"
}

RANGE_BLIND = (0, 100)
RANGE_BRIGHTNESS = (0.0, 100.0)

SUPPORT_BRIGHTNESS = 1
SUPPORT_RGB_COLOR = 2
SUPPORT_COLOR_TEMP = 4
SUPPORT_HEX_COLOR = 8

# supported features
SUPPORT_OPEN = 1
SUPPORT_CLOSE = 2
SUPPORT_SET_POSITION = 4,
SUPPORT_STOP = 8
SUPPORT_OPEN_TILT = 16
SUPPORT_CLOSE_TILT = 32
SUPPORT_STOP_TILT = 64

STATE_OPEN = "open"
STATE_OPENING = "opening"
STATE_CLOSED = "closed"
STATE_CLOSING = "closing"



"""Inels timer."""
from time import time


class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""


class pyTimer:
    """class pyTimer."""

    def __init__(self):
        """Initialize Timer."""
        self.__tick = None
        self.__elapsed_time = None
        self.__stop_after = None
        self.__start_time = None

    @property
    def tick(self):
        """Timer tick property"""
        return self.__tick

    @property
    def is_running(self):
        """Timer is running."""
        return self.__tick is not None

    @property
    def elapsed_time(self):
        """Time to stop timer."""
        return self.__elapsed_time

    def start(self, stop_after):
        """Start a new timer."""
        if self.__start_time is not None:
            # maybee only return void because its running
            raise TimerError("Timer is running. Use .stop() to stop it")

        self.__stop_after = stop_after

        self.__start_time = time()
        self.__tick = time() - self.__start_time
        self.__elapsed_time = self.__stop_after - self.__tick

    def update_tick(self):
        """Update tick time."""
        self.__tick = time() - self.__start_time
        self.__elapsed_time = self.__stop_after - self.__tick

    def stop(self):
        """Stop the timer, and report the elapsed time"""
        if self.__tick is None:
            raise TimerError("Timer is not running. Use .start() to start it")

        self.__tick = None
        self.__start_time = None
        self.__stop_after = None
        self.__elapsed_time = None
